""" A set of backends for AdaTest.
"""
import asyncio
import aiohttp
import transformers
import openai
import numpy as np
import copy
from ._filters import clean_string

class Backend():
    """ Abstract class for language model backends.
    """
    
    def __init__(self, models, sep, subsep, quote, filter_profanity):
        """ Create a new backend with the given separators and max length.
        """
        if not isinstance(models, list) and not isinstance(models, tuple):
            models = [models]
        self.model = models[0]
        self.models = models
        self.subsep = subsep
        self.sep = sep
        self.quote = quote
        self.filter_profanity = filter_profanity

        # value1_format = '"{value1}"'
        # value2_format = '"{value1}"'
        # value1_value2_format = '"{value1}" {comparator} "{value2}"'
        # value1_comparator_value2_format = '"{value1}" {comparator} "{value2}"'
        # value1_comparator_format = '"{value1}" {comparator} "{value2}"'
        # comparator_value2_format = '"{value1}" {comparator} "{value2}"'


    
    def __call__(self, prompts, max_length=None):
        """ This should be overridden by concrete subclasses.

        Parameters:
        -----------
        prompts: list of tuples, or list of lists of tuples
        """
        pass
    
    def _validate_prompts(self, prompts):
        """ Ensure the passed prompts are a list of prompt lists.
        """
        if isinstance(prompts[0][0], str):
            prompts = [prompts]
        return prompts
    
    def _varying_values(self, prompts, topic):
        """ Marks with values vary in the set of prompts.
        """

        show_topics = False
        gen_value1 = False
        gen_value2 = False
        gen_value3 = False
        for prompt in prompts:
            topics, value1s, value2s, value3s = zip(*prompt)
            show_topics = show_topics or len(set(list(topics) + [topic])) > 1
            gen_value1 = gen_value1 or len(set(value1s)) > 1
            gen_value2 = gen_value2 or len(set(value2s)) > 1
            gen_value3 = gen_value3 or len(set(value3s)) > 1
        return show_topics, gen_value1, gen_value2, gen_value3
    
    def _create_prompt_strings(self, prompts, topic):
        """ Convert prompts that are lists of tuples into strings for the LM to complete.
        """

        show_topics, gen_value1, gen_value2, gen_value3 = self._varying_values(prompts, topic)

        prompt_strings = []
        for prompt in prompts:
            topics, value1s, value2s, value3s = zip(*prompt)
            prompt_string = ""
            for p_topic, value1, value2, value3 in prompt:
                if show_topics:
                    prompt_string += self.sep + p_topic + ":" + self.sep + self.quote
                else:
                    prompt_string += self.quote
                
                if gen_value1:
                    prompt_string += value1 + self.quote
                if gen_value1 and (gen_value2 or gen_value3):
                    prompt_string += self.subsep + self.quote
                if gen_value2:
                    prompt_string += value2 + self.quote
                if gen_value2 and gen_value3:
                    prompt_string += self.subsep + self.quote
                if gen_value3:
                    prompt_string += value3 + self.quote
                prompt_string += self.sep
            if show_topics:
                prompt_strings.append(prompt_string + self.sep + topic + ":" + self.sep + self.quote)
            else:
                prompt_strings.append(prompt_string + self.quote)
        return prompt_strings
    
    def _parse_suggestion_texts(self, suggestion_texts, prompts):
        """ Parse the suggestion texts into tuples.
        """
        assert len(suggestion_texts) % len(prompts) == 0, "Missing prompt completions!"

        _, gen_value1, gen_value2, gen_value3 = self._varying_values(prompts, "") # note that "" is an unused topic argument
        
        num_samples = len(suggestion_texts) // len(prompts)
        samples = []
        for i, suggestion_text in enumerate(suggestion_texts):
            if self.filter_profanity:
                suggestion_text = clean_string(suggestion_text)
            prompt_ind = i // num_samples
            prompt = prompts[prompt_ind]
            suggestion = suggestion_text.split(self.quote+self.subsep+self.quote)
            # if len(self.quote) > 0: # strip any dangling quote
            #     suggestion[-1] = suggestion[-1][:-len(self.quote)]
            if not gen_value1:
                suggestion = [prompt[0][0]] + suggestion
            if not gen_value2:
                suggestion = suggestion[:1] + [prompt[0][2]] + suggestion[1:]
            if not gen_value3:
                suggestion = suggestion[:2] + [prompt[0][3]]

            if len(suggestion) == 3:
                samples.append(tuple(suggestion))
        return samples

           
class Transformers(Backend):
    def __init__(self, model, tokenizer, sep="\n", subsep=" ", quote="\"", filter_profanity=True):
        super().__init__(models=model, sep=sep, subsep=subsep, quote=quote)
        self.tokenizer = tokenizer
        self.device = self.model.device

        class StopAtSequence(transformers.StoppingCriteria):
            def __init__(self, stop_string, tokenizer, window_size=10):
                self.stop_string = stop_string
                self.tokenizer = tokenizer
                self.window_size = 10
                self.max_length = None
                self.prompt_length = 0
                
            def __call__(self, input_ids, scores):
                if len(input_ids[0]) > self.max_length + self.prompt_length:
                    return True

                # we need to decode rather than check the ids directly because the stop_string may get enocded differently in different contexts
                return self.tokenizer.decode(input_ids[0][-self.window_size:])[-len(self.stop_string):] == self.stop_string

        self._sep_stopper = StopAtSequence(self.quote+self.sep, self.tokenizer)
    
    def __call__(self, prompts, topic, num_samples=1, max_length=100):
        
        prompts = self._validate_prompts(prompts)
        prompt_strings = self._create_prompt_strings(prompts, topic)
        
        # monkey-patch a method that prevents the use of past_key_values
        saved_func = self.model.prepare_inputs_for_generation
        def prepare_inputs_for_generation(input_ids, **kwargs):
            if "past_key_values" in kwargs:
                return {"input_ids": input_ids, "past_key_values": kwargs["past_key_values"]}
            else:
                return {"input_ids": input_ids}
        self.model.prepare_inputs_for_generation = prepare_inputs_for_generation
        
        # run the generative LM for each prompt
        suggestion_texts = []
        for prompt_string in prompt_strings:
            input_ids = self.tokenizer.encode(prompt_string, return_tensors='pt').to(self.device)
            cache_out = self.model(input_ids[:, :-1], use_cache=True)

            for _ in range(num_samples):
                self._sep_stopper.prompt_length = 1
                self._sep_stopper.max_length = max_length
                out = self.model.sample(
                    input_ids[:, -1:], pad_token_id=self.model.config.eos_token_id,
                    stopping_criteria=self._sep_stopper,
                    past_key_values=cache_out.past_key_values # TODO: enable re-using during sample unrolling as well
                )

                # we ignore first token because it is part of the prompt
                suggestion_text = self.tokenizer.decode(out[0][1:])
                
                # we ignore the stop string to match other backends
                if suggestion_text[-len(self._sep_stopper.stop_string):] == self._sep_stopper.stop_string:
                    suggestion_text = suggestion_text[:-len(self._sep_stopper.stop_string)]
                
                suggestion_texts.append(suggestion_text)

        # restore the old function that prevents the past_key_values argument from getting passed
        self.model.prepare_inputs_for_generation = saved_func
        
        return self._parse_suggestion_texts(suggestion_texts, prompts)


class OpenAI(Backend):
    """ Backend wrapper for the OpenAI API that exposes GPT-3.
    """
    
    def __init__(self, models, api_key=None, sep="\n", subsep=" ", quote="\"", temperature=1.0, top_p=0.95, filter_profanity=True):
        super().__init__(models, sep=sep, subsep=subsep, quote=quote, filter_profanity=filter_profanity)
        self.temperature = temperature
        self.top_p = top_p
        if api_key is not None:
            openai.api_key = api_key

    def __call__(self, prompts, topic, test_type, scorer, num_samples=1, max_length=100):
        prompts = self._validate_prompts(prompts)
        # prompt_strings = self._create_prompt_strings(prompts, topic)

        # find out which values in the prompt have multiple values and so should be generated
        topics_vary, gen_value1, gen_value2, gen_value3 = self._varying_values(prompts, topic)

        # TODO: bias generation towards model failures
        # custom generation process for the "should not output" test
        # if test_type == "{} should not output {}":

        #     # create prompts to generate the inputs to the model
        #     input_prompt_strings, _, _, _ = self._create_prompt_strings(prompts, topic, prefix=[], generate=[1])

        #     # call the OpenAI API to complete the input generation prompts
        #     response = openai.Completion.create(
        #         engine=self.model, prompt=input_prompt_strings, max_tokens=max_length,
        #         temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote
        #     )
        #     input_suggestions = [choice["text"] for choice in response["choices"]]

        #     output_scores = scorer.model(input_suggestions)
        #     per_completion_token_bias_values = self._compute_bias_values(output_scores, scorer.model.output_names)

        #     # create prompts to generate the outputs to the model
        #     output_prompt_strings, _, _, _ = self._create_prompt_strings(prompts, topic, prefix=[1], generate=[2])

        #     # call the OpenAI API to complete the output generation prompts
        #     response = openai.Completion.create(
        #         engine=self.model, prompt=output_prompt_strings, max_tokens=max_length,
        #         temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote,
        #         per_completion_token_bias_values=per_completion_token_bias_values
        #     )
        #     output_suggestions = [choice["text"] for choice in response["choices"]]

        #     # then we build the final suggestions as the combination of the input and output suggestions



        # create prompts to generate the model input parameters of the tests
        prompt_strings = self._create_prompt_strings(prompts, topic)

        # see if we can stop at the first quote or need to wait for the seperator
        if gen_value1 + gen_value2 + gen_value3 == 1:
            stop_string = self.quote
        else:
            stop_string = self.quote+self.sep
        
        # call the OpenAI API to complete the prompts
        response = openai.Completion.create(
            engine=self.model, prompt=prompt_strings, max_tokens=max_length,
            temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=stop_string
        )
        suggestion_texts = [choice["text"] for choice in response["choices"]]
        
        return self._parse_suggestion_texts(suggestion_texts, prompts)

        # if self._should_generate_outputs(test_type, scorer):
        #     model_inputs = self._create_model_inputs(test_type, inputs_filled)
        #     outputs = self._generate_outputs(model_inputs, scorer)
        #     output_prompt_strings = self._create_output_prompt_strings(prompts, topic)

        #     for i in range(len(outputs)):
        #         logit_bias = {self._quote_token: 1} # we also upweight the quote token so we don't decrease the chance of ending the output
        #         for id in self._tokenizer(outputs[i])["input_ids"]:
        #             logit_bias[id] = logit_bias.get(id, 0) + 1
        #             response = openai.Completion.create(
        #                 engine=self.model, prompt=output_prompt_strings, max_tokens=max_length,
        #                 temperature=self.temperature, top_p=self.top_p, n=num_samples, stop=self.quote, logit_bias=logit_bias
        #             )
        #             suggestion_texts = [choice["text"] for choice in response["choices"]]
                






class AI21(Backend):
    """ Backend wrapper for the OpenAI API that exposes GPT-3.
    """
    
    def __init__(self, model, api_key, sep="\n", subsep=" ", quote="\"", temperature=0.95, filter_profanity=True):
        super().__init__(model, sep=sep, subsep=subsep, quote=quote, filter_profanity=filter_profanity)
        self.api_key = api_key
        self.temperature = temperature
        self.event_loop = asyncio.get_event_loop()
    
    def __call__(self, prompts, topic, num_samples=1, max_length=100):
        prompts = self._validate_prompts(prompts)
        prompt_strings = self._create_prompt_strings(prompts, topic)
        
        # define an async call to the API
        async def http_call(prompt_string):
            async with aiohttp.ClientSession() as session:
                async with session.post(f"https://api.ai21.com/studio/v1/{self.model}/complete",
                        headers={"Authorization": f"Bearer {self.api_key}"},
                        json={
                            "prompt": prompt_string, 
                            "numResults": num_samples, 
                            "maxTokens": max_length,
                            "stopSequences": [self.quote+self.sep],
                            "topKReturn": 0,
                            "temperature": self.temperature
                        }) as resp:
                    result = await resp.json()
                    return [c["data"]["text"] for c in result["completions"]]
        
        # call the AI21 API asyncronously to complete the prompts
        results = self.event_loop.run_until_complete(asyncio.gather(*[http_call(s) for s in prompt_strings]))
        suggestion_texts = []
        for result in results:
            suggestion_texts.extend(result)
        
        return self._parse_suggestion_texts(suggestion_texts, prompts)


import numpy as np
import re
import logging
import uuid
import itertools
import sentence_transformers
import openai
import scipy.stats
import transformers
import shap
from ._model import Model
from .utils import isinstance_ipython
#from transformers.tokenization_utils_base import ExplicitEnum
#from ._explorer import file_log

log = logging.getLogger(__name__)

class Scorer():
    def __new__(cls, model, *args, **kwargs):
        """ If we are wrapping an object that is already a Scorer, we just return it.
        """
        if isinstance_ipython(model, Scorer):
            return model
        else:
            return super().__new__(cls)
    
    def __init__(self, model):
        """ Auto detect the model type and subclass to the right scorer object.
        """

        # do common init stuff if this is the first time we are called
        if not hasattr(self, "_id"):
            self._id = uuid.uuid4().hex

            # ensure we have a model of type Model
            if isinstance_ipython(model, Model):
                self.model = model
            else:
                self.model = Model(model)

        # If we are in the base class we need to pick the right specialized subclass to become
        if self.__class__ is Scorer:

            # finish early if we are wrapping an object that is already a Scorer (__new__ will have already done the work)
            if isinstance_ipython(model, Scorer):
                return
            
            # see if we are scoring a generator or a classifier
            out = self.model(["string 1", "string 2"])
            if isinstance(out[0], str):
                self.__class__ = GeneratorScorer
                GeneratorScorer.__init__(self, model)
            else:
                self.__class__ = ClassifierScorer
                ClassifierScorer.__init__(self, model)
            

class DummyScorer(Scorer):
    def __init__(self):
        self._id = uuid.uuid4().hex
    def __call__(self, tests):
        out = []
        for k, test in tests.iterrows():
            try:
                score = float(test.value2)
            except:
                score = np.nan
            out.append(score)
        return np.array(out)

class ClassifierScorer(Scorer):
    """ Wraps a model and defines a callable scorer that returns a score value for any input/output pair.

    Positive scores indicate test failures, positive scores indicate tests that pass. For example if we wrap
    a text sentiment classifer the `scorer(TestTree([("this is great!", "should be", "POSITIVE")]))` will return
    a large positive value indicating that the model is very likely to correctly produce that output when given
    that input.
    """

    def __init__(self, model, topk=1, output_names=None, method="dirichlet", dirichlet_concentration=10):
        """ Create a new scorer given a model that returns a probability vector for each input string.
        
        Parameters:
        -----------
        model : callable
            A model that is callable with a single argument (which is a list of strings) and returns a matrix of outputs.

        topk : int
            The number of top outputs to consider when scoring tests. For example topk=2 causes "should not be" tests to
            check the top two model outputs.

        output_names : list of strings
            A list of strings that correspond to the outputs of the model. If None, model.output_names is used.

        method : 'margin' or 'dirichlet'
            The scoring method to use. Dirichlet is preferred, but margin is available for backwards compatibility.

        dirichlet_concentration : float
            The concentration parameter for the dirichlet scoring method. It is in the units o pseudo-counts where larger
            values lead to a tighter prior centered around the model's probability outputs (so scores are more likely to
            be -1 or +1).
        """
        super().__init__(model)

        # we don't want to re-init a class if init has alrady been done (this can happen when Scorer(maybe_scorer) is called)
        if hasattr(self, "supported_test_types"):
            return # already initialized

        self.supported_test_types = [
            "{} should output {}",
            "{} should not output {}",
            "{} should have the same output as {}",
            "{} should not be more {} than {}",
            #"{} should be invertable. [] is the output.",
            #"{} should not output text containing {}",
            #"{} should output toxic text. [] is the output.",
            #"{} should not output toxic text. [] is the output."
        ]
        # {} should output {}
        # {} should not output {}
        # {} should have the same output as {}
        # {} should not be more {} than {}
        # {} should be invertable. [] is the output.
        # {} should not output text containing {}
        # {} should output toxic text. [] is the output.
        # {} should not output toxic text. [] is the output.

        # extract output names from the model if they are not provided directly
        if output_names is None:
            self.output_names = self.model.output_names
        else:
            self.output_names = output_names
        
        if not callable(self.output_names):
            self._output_name_to_index = {v: i for i, v in enumerate(self.output_names)}
        self.topk = topk
        self.output_type = "classification"
        self.method = method
        self.dirichlet_concentration = dirichlet_concentration

    def __call__(self, tests, output_bias=0.5):
        """ Score a set of tests.

        Parameters
        ----------
        tests : pandas.DataFrame
            A dataframe of tests.

        output_bias : float
            How much to bias the output entries in a test towards the actual output from the model.
        """
        if self.output_type == "classification":
            eval_inputs = []
            eval_inds = []
            variations1 = []
            variations2 = []
            for i, (k, test) in enumerate(tests.iterrows()):
                if test.type == "{} should not output {}" or test.type == "{} should output {}":
                    v1 = expand_template(test.value1)
                    for s1 in v1:
                        eval_inputs.append(s1)
                        eval_inds.append(i)
                    variations1.append(v1)
                    variations2.append(None)
                elif test.type == "{} should have the same output as {}":
                    # eval_inputs.append(test.value1)
                    # eval_inputs.append(test.value2)
                    v1 = expand_template(test.value1)
                    v2 = expand_template(test.value2)
                    for s1 in v1:
                        for s2 in v2:
                            eval_inputs.append(s1)
                            eval_inputs.append(s2)
                            eval_inds.append(i)
                            eval_inds.append(i)
                    variations1.append(v1)
                    variations2.append(v2)

            try:
                model_out = self.model(eval_inputs)
            except Exception as e:
                model_out = np.zeros((len(eval_inputs), len(self.model.output_names))) * np.nan # TODO: remove this hack after the user study
                log.error(e)
                log.error(eval_inputs)
                log.error("The model threw an exception when evaluating inputs! We are patching this disaster with np.nan for the sake of the user study!")

            out = [[] for _ in range(tests.shape[0])]
            out_pos = 0
            i = 0
            value1_outputs = [{} for _ in range(tests.shape[0])]
            value2_outputs = [{} for _ in range(tests.shape[0])]
            while i < len(model_out):
                out_pos = eval_inds[i]

                test_type = tests.iloc[out_pos]["type"]
                if test_type == "{} should not output {}" or test_type == "{} should output {}":

                    # save the top model outputs
                    inds = np.argsort(-model_out[i])
                    shown_tmp = {}
                    for j in inds[:5]:
                        shown_tmp[self.model.output_names[j]] = float(model_out[i][j])
                    value1_outputs[out_pos] = shown_tmp

                    # TODO: here we need to not assume the LM genreates the output, but that it just generates the inputs
                    # then we can embed the tests for each of the top 10? 100? outputs and rank them by how well their
                    # embeddings match the test embeddings for the prompt
                    token_to_check = tests.iloc[out_pos]['value2']

                    # To make sharing test trees between models slightly easier we fall back to looking for different
                    # capitalizations of the output if the original one doesn't exist
                    if token_to_check not in self._output_name_to_index:
                        if token_to_check.capitalize() in self._output_name_to_index:
                            token_to_check = token_to_check.capitalize()
                        elif token_to_check.lower() in self._output_name_to_index:
                            token_to_check = token_to_check.lower()
                    
                    # multiple tokens can be checked at the same time with templates
                    out_val = np.nan
                    for token_part in expand_template(token_to_check):
                        ind = self._output_name_to_index.get(token_part, None)
                        if ind is not None and model_out[i] is not None:
                            sorted_values = np.argsort(model_out[i])
                            topk = topk_threshold_ind(ind, sorted_values, self.topk)

                            if self.method == "dirichlet":
                                raw_score = compute_dirichlet_score(
                                    ind, model_out[i], self.topk,
                                    concentration=self.dirichlet_concentration,
                                    # we treat values less than 10% of the topk value as unlikely to impact the results
                                    # this is used to avoid unnecessary computation
                                    domination_threshold=model_out[i][topk] / 10 
                                )

                                if test_type == "{} should output {}":
                                    score = raw_score
                                else:
                                    score = -raw_score

                            if np.isnan(model_out[i][ind]):
                                score = np.nan
                            elif model_out[i][ind] > model_out[i][topk]:
                                if self.method == "dirichlet":
                                    # mval = 1 / len(model_out[i]) if self.topk == 1 else 0 # minimum value possible while being at the top
                                    # score = (raw_score - mval) / (1 - mval) # scale from 0 to 1
                                    score = raw_score
                                else:
                                    score = model_out[i][ind] - model_out[i][topk]
                            else:
                                if self.method == "dirichlet":
                                    # mval = 1 / (self.topk + 1) # maximum value possible while not being at the top
                                    # score = (raw_score - 1 + mval) / (1 - mval) # scale from 0 to 1
                                    score = raw_score - 1
                                else:
                                    mask = (model_out[i] <= model_out[i][topk]) & (model_out[i] > model_out[i][ind])
                                    score = (model_out[i][ind] - model_out[i][mask]).sum()
                            if test_type == "{} should output {}":
                                score *= -1
                            # out_val = max(score, out_val)
                            out[out_pos].append(score)
                    # out[out_pos] = max(out[out_pos], out_val)
                    i += 1
                elif test_type == "{} should have the same output as {}":

                    # save the top model outputs
                    inds = np.argsort(-model_out[i])
                    shown_tmp = {}
                    for j in inds[:5]:
                        shown_tmp[self.model.output_names[j]] = float(model_out[i][j])
                    value1_outputs[out_pos] = shown_tmp
                    inds = np.argsort(-model_out[i+1])
                    shown_tmp = {}
                    for j in inds[:5]:
                        shown_tmp[self.model.output_names[j]] = float(model_out[i+1][j])
                    value2_outputs[out_pos] = shown_tmp
                    
                    if np.isnan(model_out[i][0]):
                        score = np.nan
                    elif self.method == "dirichlet":
                        score = compute_dirichlet_equality_score(
                            model_out[i], model_out[i+1], self.topk, concentration=self.dirichlet_concentration,
                            # we treat values less than 1% of the top value as unlikely to impact the results
                            # this is used to avoid unnecessary computation (note this is used more aggressivly than for the non-equality score)
                            domination_threshold=min(np.max(model_out[i]), np.max(model_out[i+1])) / 100
                        )
                    else:
                        score = equality_score(model_out[i], model_out[i+1])
                    # out[out_pos] = max(out[out_pos], score)
                    out[out_pos].append(score)
                    i += 2
                else:
                    raise Exception(f"Test type '{test_type}' not yet supported!")

                # out_pos += 1
            return {
                "scores": out,
                "value1_outputs": value1_outputs,
                "value2_outputs": value2_outputs
            }
        else:
            raise Exception(f"Output type {self.output_type} not yet supported!")

    def suggest_outputs(self, current, num_suggestions=20):
        prompt = ""
        for c in current:
            prompt += '"'+c+'"\n'
        prompt += '"{output}'
        response = openai.Completion.create(
            engine='curie-instruct-beta', prompt=[prompt.format(output=o) for o in self.output_names], max_tokens=0, # self.engine
            temperature=0, n=1, stop='\"', logprobs=0, echo=True
        )
        lines = [sum(choice["logprobs"]["token_logprobs"][11:]) for choice in response["choices"]]
        pairs = list([v for v in zip(lines, self.output_names) if v[1] not in current])
        pairs.sort()
        return [v[1] for v in list(reversed(pairs))[:num_suggestions]]


class GeneratorScorer(Scorer):
    """ Wraps a text generation model in a scorer that can score the target model against tests.
    """

    def __init__(self, model, reverse_model=None, embedding_model=None, feature_models={}, similarity_threshold=0.9):
        """ Initializes a new scorer for a given target model.

        Parameters
        ----------
        model : callable
            The model we are scoring against the tests. It is expected to be a function that takes a list of strings as
            input and returns a list of strings as output.

        reverse_model : callable
            The inverse of model, such that x = reverse_model(model(x)). Note that this is optional and only required
            to use the 'should be invertable' test type. If for example model is an EN to ES translation model, then
            reverse_model should be an ES to EN translation model. Just like model, reverse_model is expected to take a
            list of strings as input and return a list of strings as output.

        embedding_model : callable
            Used to embed the tests for comparison purposes. TODO: shouldn't this be shared with the main adatest backend?

        feature_models : dict
            A dictionary of classifiers that can be used to score the generated output. The keys are the names of the
            classifier label that is predicted and the values are the classifiers themselves. Each classifer is expected
            to take a list of strings as input and return a list of floats between 0 and 1.

        similarity_threshold : float
            The threshold for the similarity between the generated output and the expected output. If set to 1.0, then
            exact string matching is performed. If < 1.0 then this is the threshold on the sementantic similarity (a cos
            similarity score between 0 and 1).
            TODO: this was just used for the 'should be invertable' test type. Should we use it for other test types as well?
        """
        super().__init__(model)

        # we don't want to re-init a class if init has alrady been done (this can happen when Scorer(maybe_scorer) is called)
        if hasattr(self, "supported_test_types"):
            return # already initialized

        self.reverse_model = reverse_model
        self.embedding_model = embedding_model
        self.similarity_threshold = similarity_threshold
        self.feature_models = feature_models

        self.supported_test_types = [
            "{} should output {}",
            "{} should not output {}", # TODO: should this use semantic similarity instead of exact string match?
            "{} should have the same output as {}",
            "{} should output text containing {}",
            "{} should not output text containing {}",
            "{} should output text that is {}",
            "{} should not output text that is {}",
            "{} should not be completed to become {}"
        ]
        # '"{}[]" should not become "{}" when completed. 

        # see if we can support the inversion test
        if self.reverse_model is not None and self.embedding_model is not None: # TODO: do we need an embeddig model or just use the backend?
            self.supported_test_types.append("{} should be invertable. [] is the output.") # Note that {} means user-editable, [] means read-only

        # # see if we have user-provided classifier tests
        # for k in self.feature_models:
        #     self.supported_test_types.extend([
        #         "{} should output "+k+" text. [] is the output.",
        #         "{} should not output "+k+" text. [] is the output."
        #     ])

        

    def __call__(self, tests):

        # run the model on the inputs
        eval_inputs = []
        eval_inds = []
        eval_reverse_pos = []
        variations1 = []
        variations2 = []
        eval_output_feature_pos = {k: [] for k in self.feature_models}
        eval_input_feature_pos = {k: [] for k in self.feature_models}
        eval_io_feature_pos = {k: [] for k in self.feature_models}
        for i, (k, test) in enumerate(tests.iterrows()):
            if test.type in ["{} should not output {}", "{} should output {}",
                             "{} should not output text containing {}", "{} should output text containing {}"]:
                v1 = expand_template(test.value1)
                for s1 in v1:
                    eval_inputs.append(s1)
                    eval_inds.append(i)
                variations1.append(v1)
                variations2.append(None)
            elif test.type == "{} should not output text that is {}" or test.type == "{} should output text that is {}":
                v1 = expand_template(test.value1)
                v2 = expand_template(test.value2)
                for s1 in v1:
                    eval_inputs.append(s1)
                    eval_inds.append(i)
                    for s2 in v2: # mark each feature in the template for evaluation
                        if s2 in eval_output_feature_pos:
                            eval_output_feature_pos[s2].append(len(eval_inputs)-1)
                variations1.append(v1)
                variations2.append(None)
            elif test.type == "{} should not be completed to become {}":
                v1 = expand_template(test.value1)
                v2 = expand_template(test.value2)
                for s1 in v1:
                    eval_inputs.append(s1)
                    eval_inds.append(i)
                    for s2 in v2: # mark each feature in the template for evaluation
                        if s2 in eval_input_feature_pos:
                            eval_io_feature_pos[s2].append(len(eval_inputs)-1)
                            eval_input_feature_pos[s2].append(len(eval_inputs)-1)
                variations1.append(v1)
                variations2.append(None)
            elif test.type == "{} should be invertable.":
                v1 = expand_template(test.value1)
                for s1 in v1:
                    eval_inputs.append(s1)
                    eval_inds.append(i)
                    eval_reverse_pos.append(len(eval_inputs) - 1)
            elif test.type == "{} should have the same output as {}":
                v1 = expand_template(test.value1)
                v2 = expand_template(test.value2)
                for s1 in v1:
                    for s2 in v2:
                        eval_inputs.append(s1)
                        eval_inputs.append(s2)
                        eval_inds.append(i)
                        eval_inds.append(i)
                variations1.append(v1)
                variations2.append(v2)
        try:
            # model_out = self.model(eval_inputs)
            model_out = self.model(eval_inputs, completions=self.completions) # now a list of lists
        except Exception as e:
            model_out = ["ERROR" for _ in range(len(eval_inputs))]#np.zeros((len(eval_inputs), len(self.model.output_names))) * np.nan # TODO: remove this hack after the user study
            log.error(e)
            log.error(eval_inputs)
            log.error("The model threw an exception when evaluating inputs! We are patching this disaster with 'ERROR' for the sake of the user study!")

        # run feature models when we need to
        model_output_feature = {}
        model_input_feature = {}
        model_io_feature = {}
        for k in self.feature_models:
            if len(eval_output_feature_pos[k]) > 0:
                feature_out = self.feature_models[k]([model_out[ind] for ind in eval_output_feature_pos[k]])
                model_output_feature[k] = [None for _ in model_out]
                for i, ind in enumerate(eval_output_feature_pos[k]):
                    model_output_feature[k][ind] = feature_out[i]
            if len(eval_input_feature_pos[k]) > 0:
                feature_out = self.feature_models[k]([eval_inputs[ind] for ind in eval_input_feature_pos[k]])
                model_input_feature[k] = [None for _ in model_out]
                for i, ind in enumerate(eval_input_feature_pos[k]):
                    model_input_feature[k][ind] = feature_out[i]
            if len(eval_io_feature_pos[k]) > 0:
                feature_out = self.feature_models[k]([eval_inputs[ind] + model_out[ind] for ind in eval_io_feature_pos[k]])
                model_io_feature[k] = [None for _ in model_out]
                for i, ind in enumerate(eval_io_feature_pos[k]):
                    model_io_feature[k][ind] = feature_out[i]

        # feature_model_inputs = {k: [] for k in self.feature_models}
        # i = 0
        # while i < len(model_out):
        #     if test.type == "{} should not output text that is {}" or test.type == "{} should output text that is {}":
        #         if test.value1 in self.feature_models:
        #             feature_model_inputs[test.value1].append((i, model_out[i]))
        #     i += 1

        # run the reverse model on any outputs we need to
        # eval_reverse_inputs = []
        
        # for i, (k, test) in enumerate(tests.iterrows()):
        #     if test.comparator == "should be invertable.":
        #         v1 = expand_template(test.value1)
        #         for s1 in v1:
        #             eval_reverse_inputs.append(s1)
        #             eval_reverse_inds.append(i)
        if len(eval_reverse_pos) > 0:
            model_reverse_out = [None for _ in model_out]
            input_embed = [None for _ in model_out]
            round_trip_embed = [None for _ in model_out]
            try:
                # compute input embedding
                tmp = self.embedding_model.encode([eval_inputs[ind] for ind in eval_reverse_pos], convert_to_tensor=True, show_progress_bar=False).cpu()
                for i, ind in enumerate(eval_reverse_pos):
                    input_embed[ind] = tmp[i]

                # compute reverse model output
                reverse_out = self.reverse_model([model_out[ind] for ind in eval_reverse_pos])
                for i, ind in enumerate(eval_reverse_pos):
                    model_reverse_out[ind] = str(reverse_out[i])

                # compute round trip embedding
                tmp = self.embedding_model.encode(reverse_out, convert_to_tensor=True, show_progress_bar=False).cpu()
                for i, ind in enumerate(eval_reverse_pos):
                    round_trip_embed[ind] = tmp[i]

            except Exception as e:
                model_reverse_out = ["ERROR" for _ in range(len(model_out))]
                log.error(e)
                log.error("The reverse model threw an exception when evaluating inputs! We are patching this disaster with 'ERROR' for the sake of the user study!")
        else:
            model_reverse_out = []

        out = [[] for _ in range(tests.shape[0])]
        out_pos = 0
        i = 0
        value1_outputs = [{} for _ in range(tests.shape[0])]
        value2_outputs = [{} for _ in range(tests.shape[0])]
        while i < len(model_out):
            out_pos = eval_inds[i]

            test_type = tests.iloc[out_pos]["type"]
            if test_type == "{} should not output {}" or test_type == "{} should output {}" or test_type == "{} should not output text containing {}":
                invert = -1 if test_type == "{} should not output {}" or test_type == "{} should not output text containing {}" else 1
                contain_check = test_type == "{} should not output text containing {}" or test_type == "{} should output text containing {}"
                
                # auto fill missing outputs
                if tests.iloc[out_pos]['value2'] is None:
                    tests.loc[tests.index[out_pos], 'value2'] = str(model_out[i])
                
                # save the model output
                value1_outputs[out_pos]  = {}
                value1_outputs[out_pos][model_out[i]] = 1

                # multiple tokens can be checked at the same time with templates
                for token_part in expand_template(tests.iloc[out_pos]['value2']):
                    if contain_check:
                        out[out_pos].append(-invert if token_part in model_out[i] else invert)
                    else:
                        out[out_pos].append(-invert if model_out[i] == token_part else invert)
                i += 1
            elif test_type == "{} should output text that is {}" or test_type == "{} should not output text that is {}":
                should_be = test_type == "{} should output text that is {}"

                # multiple features can be checked at the same time with templates
                for feature in expand_template(tests.iloc[out_pos]['value2']):
                    if feature in model_output_feature:
                        feature_prob = model_output_feature[feature][i]
                        if should_be:
                            score = (0.5 - feature_prob) * 2
                        else:
                            score = (feature_prob - 0.5) * 2
                    else:
                        score = np.nan
                    out[out_pos].append(score)
                
                # save the model round trip output
                value1_outputs[out_pos]  = {}
                value1_outputs[out_pos][model_out[i]] = 1

                i += 1
            
            elif test_type == "{} should not be completed to become {}":

                # multiple features can be checked at the same time with templates
                for feature in expand_template(tests.iloc[out_pos]['value2']):
                    if feature in model_io_feature:# and feature in model_input_feature is implied
                        io_feature_prob = model_io_feature[feature][i]
                        input_feature_prob = model_input_feature[feature][i]
                        if input_feature_prob <= 0.5 and io_feature_prob >= 0.5:
                            score = io_feature_prob - input_feature_prob
                        elif input_feature_prob <= 0.5 and io_feature_prob <= 0.5:
                            score = io_feature_prob - input_feature_prob - 0.5
                        elif input_feature_prob >= 0.5 and io_feature_prob >= 0.5:
                            score = io_feature_prob - input_feature_prob + 0.5
                        else: # if input_feature_prob >= 0.5 and io_feature_prob <= 0.5
                            score = io_feature_prob - input_feature_prob
                    else:
                        score = np.nan
                    out[out_pos].append(score)
                
                # save the model output
                value1_outputs[out_pos] = {}
                value1_outputs[out_pos]["string"] = model_out[i]

                value2_outputs[out_pos] = {}
                value2_outputs[out_pos]["input prob"] = input_feature_prob
                value2_outputs[out_pos]["input+output prob"] = io_feature_prob

                i += 1
            
            elif test_type == "{} should be invertable.":
                
                # compare embedding distances
                score = sentence_transformers.util.pytorch_cos_sim(input_embed[i], round_trip_embed[i]).numpy()[0][0]
                out[out_pos].append(self.similarity_threshold-score)

                # update the output since it is always computed in inversion tests
                tests.loc[tests.index[out_pos], 'value2'] = str(model_reverse_out[i])
                
                # save the model round trip output
                value1_outputs[out_pos]  = {}
                value1_outputs[out_pos][str(model_out[i])] = 1

                i += 1
            elif test_type == "{} should have the same output as {}":

                # save the model outputs
                value1_outputs[out_pos]  = {}
                value1_outputs[out_pos][model_out[i]] = 1
                value2_outputs[out_pos]  = {}
                value2_outputs[out_pos][model_out[i+1]] = 1
                
                # save the score
                out[out_pos].append(1 if model_out[i] == model_out[i+1] else -1)
                i += 2
            else:
                raise Exception(f"Test type type '{test_type}' not yet supported!")

            # out_pos += 1
        return {
            "scores": out,
            "value1_outputs": value1_outputs,
            "value2_outputs": value2_outputs
        }


def expand_template(s):
    """ Expand a template string into a list of strings.
    """
    # parts = []
    # for s in strings:
    matches = re.findall("{[^}]*}", s)
    s = re.sub("{[^}]*}", "{}", s)
    template_groups = [str(m)[1:-1].split("|") for m in matches]
    try:
        return [s.format(*parts) for parts in itertools.product(*template_groups)]
    except ValueError:
        return [s] # we return the template not filled in if it is invalid

def clean_template(s):
    """ This removes duplicate template entries.
    """
    matches = re.findall("{[^}]*}", s)
    s = re.sub("{[^}]*}", "{}", s)
    template_groups = [str(m)[1:-1].split("|") for m in matches]
    clean_groups = ["{"+"|".join(list({v: None for v in g}.keys()))+"}" for g in template_groups]
    try:
        return s.format(*clean_groups)
    except ValueError:
        return s # we return the template not cleaned in if it is invalid

def topk_threshold_ind(ind, sorted_values, k):
    """ Return the threshold value for which if ind dropped below it would not be in the top k (without other scores changing).
    """
    if ind in sorted_values[-k:]:
        topk = sorted_values[-k - 1]
    else:
        topk = sorted_values[-k]
    if topk == ind:
        topk = sorted_values[-k - 1]
    return topk

def compute_dirichlet_score(ind, model_output, k=1, concentration=10, empirical_samples=1000, domination_threshold=1e-5):
    """ Compute the probability that ind is in the top k set of probabilities.

    This is done by sampling from a Dirichlet distribution with concentration parameter centered around the model output
    (by default assuming a concentration weight of 10 pseudo-counts).

    Parameters
    ----------
    ind : int
        The index to compute the score for.

    model_output : np.array
        The model output probabilities.

    k : int
        The number of top k probabilities to consider ind in.

    concentration : float
        The concentration parameter for the Dirichlet distribution. Larger values make the distribution have lower variance.

    empirical_samples : int
        We can't calculate the probability of ind being in the top k set of probabilities exactly, so we sample from the
        distribution to get an empirical estimate. This controls the number of samples to take.

    domination_threshold : float
        Below this value we assume that these output dims can be safely ignored. This is used to avoid unnessecary computation.
    """

    # a small heuristic tie breaker to allow better ranking of small effects
    # this helps when the effect sizes are so small the empirical_samples can't capture the differences
    tie_breaker = model_output[ind]*1e-4
    
    # if our output of interest is dominated then we just skip the work and return essentially 0
    if model_output[ind] < domination_threshold:
        return tie_breaker
    
    # shrink the number of dims we have to deal with by collapsing low probability dims
    bundles = []
    bundle = []
    bundle_sizes = []
    inds = np.argsort(model_output)
    bundle_size = 0
    new_ind = -1
    for i,sind in enumerate(inds):
        if bundle_size + model_output[sind] < domination_threshold:
            bundle.append(sind)
            bundle_size += model_output[sind]
        else:
            if len(bundle) > 0:
                bundles.append(bundle)
                bundle_sizes.append(bundle_size)

            if sind == ind:
                new_ind = len(bundles)
            bundle = [sind]
            bundle_size = model_output[sind]
    bundles.append(bundle)
    bundle_sizes.append(bundle_size)
    
    # normalize the scores for the Dirichlet parameter
    normed_output = np.array(bundle_sizes) + 1e-6
    normed_output /= normed_output.sum()
    normed_output *= concentration
    
    if k == 1:
        sort_inds = np.argmax(scipy.stats.dirichlet.rvs(normed_output, empirical_samples, random_state=0), 1)
        return min(1, (sort_inds == new_ind).mean() + tie_breaker)
    else:
        sort_inds = np.argsort(-scipy.stats.dirichlet.rvs(normed_output, empirical_samples, random_state=0), 1)
        return min(1, ((sort_inds[:,:k] - new_ind) == 0).sum() / sort_inds.shape[0] + tie_breaker)

def compute_dirichlet_equality_score(model_output1, model_output2, k=1, concentration=10, empirical_samples=1000, domination_threshold=1e-5):
    """ Compute the probability that ind is in the top k set of probabilities.

    This is done by sampling from a Dirichlet distribution with concentration parameter centered around the model output
    and assuming a concentration weight of 10 pseudo-counts.

    Parameters
    ----------
    ind : int
        The index to compute the score for.

    model_output : np.array
        The model output probabilities.

    k : int
        The number of top k probabilities to consider ind in.

    concentration : float
        The concentration parameter for the Dirichlet distribution. Larger values make the distribution have lower variance.

    empirical_samples : int
        We can't calculate the probability of ind being in the top k set of probabilities exactly, so we sample from the
        distribution to get an empirical estimate. This controls the number of samples to take.

    domination_threshold : float
        Below this value we assume that these output dims can be safely ignored. This is used to avoid unnessecary computation.
    """

    assert len(model_output1) == len(model_output2)

    # shrink the number of dims we have to deal with by collapsing low probability dims
    used_inds = [i for i in range(len(model_output1)) if model_output1[i] > domination_threshold or model_output2[i] > domination_threshold]
    # model_output1 = model_output1[used_inds]
    # model_output2 = model_output2[used_inds]
    model_output1_padded = np.zeros(len(used_inds) + 1)
    model_output1_padded[1:] = model_output1[used_inds]
    model_output1_padded[0] = 1 - np.sum(model_output1)
    model_output2_padded = np.zeros(len(used_inds) + 1)
    model_output2_padded[1:] = model_output2[used_inds]
    model_output2_padded[0] = 1 - np.sum(model_output2)

    assert model_output1_padded[0] >= -1e-6 and model_output2_padded[0] >= -1e-6, "The given model output probabilities do not sum to 1!"
    model_output1_padded[0] = max(model_output1_padded[0], 0)
    model_output2_padded[0] = max(model_output2_padded[0], 0)

    # normalize the scores for the Dirichlet parameter
    normed_output1 = np.array(model_output1_padded) + 1e-6
    normed_output1 /= normed_output1.sum()
    normed_output1 *= concentration
    normed_output2 = np.array(model_output2_padded) + 1e-6
    normed_output2 /= normed_output2.sum()
    normed_output2 *= concentration
    
    if k == 1:
        sort_inds1 = np.argmax(scipy.stats.dirichlet.rvs(normed_output1, empirical_samples, random_state=0), 1)
        sort_inds2 = np.argmax(scipy.stats.dirichlet.rvs(normed_output2, empirical_samples, random_state=0), 1)

        # the average number of matches, excluding the first position (which is a bucket for all dominated, low prob, dims)
        match_rate = ((sort_inds1 - sort_inds2 == 0) * (sort_inds1 != 0)).mean()

        if np.argmax(model_output1) == np.argmax(model_output2):
            return -match_rate
        else:
            return 1 - match_rate
    else:
        raise Exception("The 'should be the same as for' is not implemented for topk > 1!")

def equality_score(output_values1, output_values2, topk=1):
    assert topk == 1
    ind1 = np.argmax(output_values1)
    ind2 = np.argmax(output_values2)
    max1 = output_values1[ind1]
    max2 = output_values2[ind2]
    margins = np.zeros(len(output_values1))

    if ind1 != ind2:
        min_margin = 1e6
        for i in range(len(output_values1)):
            score1 = max(0, max1 - output_values1[i])
            score2 = max(0, max2 - output_values2[i])
            margin = score1 + score2
            if margin < min_margin:
                min_margin = margin
        return min_margin
    else:
        val1 = output_values1[ind1]
        output_values1[ind1] = np.nan
        score1 = val1 - np.nanmax(output_values1)
        output_values1[ind1] = val1

        val2 = output_values2[ind2]
        output_values2[ind2] = np.nan
        score2 = val2 - np.nanmax(output_values2)
        output_values2[ind2] = val2
        return -min(score1, score2)



import re

def parse_test_type(test_type):
    part_names = ["text1", "value1", "text2", "value2", "text3", "value3", "text4"]
    parts = re.split(r"(\{\}|\[\])", test_type)
    part_values = ["" for _ in range(7)]
    for i, part in enumerate(parts):
        part_values[i] = part
    return {name: value for name,value in zip(part_names, part_values)}

# https://codereview.stackexchange.com/questions/253198/improved-isinstance-for-ipython
def isinstance_ipython(obj, ref_class):
    def _class_name(obj):
        name = getattr(obj, '__qualname__', getattr(obj, '__name__', ''))
        return (getattr(obj, '__module__', '') + '.' + name).lstrip('.')
    return (isinstance(obj, ref_class) or 
            _class_name(type(obj)) == _class_name(ref_class))


import pandas as pd
import uuid
import logging
import os
import io
from ._prompt_builder import PromptBuilder
from ._test_tree_browser import TestTreeBrowser, is_subtopic

class TestTree():
    """ A hierarchically organized set of tests represented as a DataFrame.

    This represents a hierarchically organized set of tests that all target a specific class of models (such as sentiment
    analysis models, or translation models). To interact with a test tree you can use either the `__call__` method to
    view and create tests directly in a Jupyter notebook, or you can call the `serve` method to launch a standalone
    webserver. A TestTree object also conforms to most of the standard pandas DataFrame API.
    """

    def __init__(self, tests=None, index=None, **kwargs):
        """ Create a new test tree.

        Parameters
        ----------
        tests : str or DataFrame or list or None
            The tests to load as a test tree. If a string is provided, it is assumed to be a path to a CSV file containing
            the tests. Otherwise tests is passed to the pandas DataFrame constructor to load the tests as a DataFrame.

        index : list or list-like or None
            Assigns an index to underlying tests frame, or auto generates if not provided.

        kwargs : dict
            Additional keyword arguments are passed to the pandas DataFrame constructor.
        """

        # the canonical ordered list of test tree columns
        column_names = ['topic', 'type', 'value1', 'value2', 'value3', 'author', 'description']

        # create a new test tree in memory
        if tests is None:
            self._tests = pd.DataFrame([], columns=column_names)
            self._tests_location = None

        # create a new test tree on disk (lazily saved)
        elif isinstance(tests, str) and not os.path.isfile(tests):
            self._tests = pd.DataFrame([], columns=column_names)
            self._tests_location = tests

        # load the test tree from a file or IO stream
        elif isinstance(tests, str) or isinstance(tests, io.TextIOBase):
            self._tests_location = tests
            if os.path.isfile(tests) or isinstance(tests, io.TextIOBase):
                self._tests = pd.read_csv(tests, index_col=0, dtype={
                    "topic": str, "type": str, "value1": str, "value2": str,
                    "value3": str, "author": str, "description": str
                    }, keep_default_na=False
                )
            else:
                raise Exception(f"The provided tests file does not exist: {tests}. If you wish to create a new file use `auto_save=True`")

        else:
            if index is None:
                index = [uuid.uuid4().hex for _ in range(len(tests))]
            self._tests = pd.DataFrame(tests, **kwargs)
            self._tests.index = index
            self._tests_location = None

        # # ensure auto saving is possible when requested
        # if auto_save and self._tests_location is None:
        #     raise Exception("auto_save=True is only supported when loading from a file or IO stream")
        # self.auto_save = auto_save

        # ensure we at least have a type column
        if "type" not in self._tests.columns:
            raise Exception("The test tree being loaded must contain a 'type' column!")

        # fill in any other missing columns
        for column in ["topic", "value1", "value2", "value3", "description"]:
            if column not in self._tests.columns:
                self._tests[column] = ["" for _ in range(self._tests.shape[0])]
        if "author" not in self._tests.columns:
            self._tests["author"] = ["anonymous" for _ in range(self._tests.shape[0])]

        # ensure that all topics have a topic_marker entry
        marked_topics = {t: True for t in set(self._tests.loc[self._tests["type"] == "topic_marker"]["topic"])}
        for topic in set(self._tests["topic"]):
            parts = topic.split("/")
            for i in range(1, len(parts)+1):
                parent_topic = "/".join(parts[:i])
                if parent_topic not in marked_topics:
                    self._tests.loc[uuid.uuid4().hex] = {
                        "type": "topic_marker",
                        "topic": parent_topic,
                        "author": "anonymous",
                        "value1": "",
                        "value2": "",
                        "value3": "",
                        "description": ""
                    }
                    marked_topics[parent_topic] = True

        # drop any duplicate index values
        self._tests = self._tests.groupby(level=0).first()

        # drop any duplicate rows
        self._tests.drop_duplicates(["topic", "type", "value1", "value2", "value3"], inplace=True)

        # put the columns in a consistent order
        self._tests = self._tests[column_names + [c for c in self._tests.columns if c not in column_names]]

        # # keep track of our original state
        # if self.auto_save:
        #     self._last_saved_tests = self._tests.copy()

    def __getitem__(self, key):
        """ TestSets act just like a DataFrame when sliced. """
        subset = self._tests[key]
        if hasattr(subset, 'columns') and len(set(["type", "topic", "value1", "value2", "value3"]) - set(subset.columns)) == 0:
            return self.__class__(subset, index=subset.index)
        return subset

    def __setitem__(self, key, value):
        """ TestSets act just like a DataFrame when sliced, including assignment. """
        self._tests[key] = value

    # all these methods directly expose the underlying DataFrame API
    @property
    def loc(self):
        return TestTreeLocIndexer(self)
    @property
    def iloc(self):
        return TestTreeILocIndexer(self)
    @property
    def index(self):
        return self._tests.index
    @property
    def columns(self):
        return self._tests.columns
    @property
    def shape(self):
        return self._tests.shape
    @property
    def iterrows(self):
        return self._tests.iterrows
    @property
    def groupby(self):
        return self._tests.groupby
    @property
    def drop(self):
        return self._tests.drop
    @property
    def insert(self):
        return self._tests.insert
    def __len__(self):
        return self._tests.__len__()
    def __setitem__(self, key, value):
        return self._tests.__setitem__(key, value)
    def to_csv(self, file=None):
        if file is None:
            self._tests.to_csv(self._tests_location)
        else:
            self._tests.to_csv(file)

    def topic(self, topic):
        """ Return a subset of the test tree containing only tests that match the given topic.

        Parameters
        ----------
        topic : str
            The topic to filter the test tree by.
        """
        ids = [id for id, test in self._tests.iterrows() if is_subtopic(topic, test.topic)]
        return self.loc[ids]

    def __call__(self, scorer=None, dataset=None, auto_save=False, user="anonymous", recompute_scores=False, drop_inactive_score_columns=False,
                 max_suggestions=100, suggestion_thread_budget=0.5, prompt_builder=PromptBuilder(), active_backend="default", starting_path="",
                 embedding_model=None, score_filter=-1e10, topic_model_scale=0):
        """ Apply this test tree to a scorer/model and browse/edit the tests to adapt them to the target model.

        Applying a test tree to a target model (wrapped by a scorer) creates a TestTreeBrowser object that can be used to
        browse the tree and add new tests to adapt it to the target model.
        
        Parameters
        ----------
        scorer : adatest.Scorer or callable
            The scorer (that wraps a model) to used to score the tests. If a function is provided, it will be wrapped in a scorer.
            Passing a dictionary of scorers will score multiple models at the same time. Note that the models are expected to take
            a list of strings as input, and output either a classification probability vector or a string.

        dataset : adatest.Dataset
            A dataset to use when suggesting new tests for the test tree.

        auto_save : bool
            Whether to automatically save the test tree after each edit.

        user : str
            The user name to author new tests with.

        recompute_scores : bool
            Whether to recompute the scores of the tests that already have score values in the test tree.

        drop_inactive_score_columns : bool
            Whether to drop the score columns in the test tree that do not match any of the passed scorers.

        max_suggestions : int
            The maximum number of suggestions to generate each time the user asks for test suggestions.

        suggestion_thread_budget : float
            This controls how many parallel suggestion processes to use when generating suggestions. A value of 0 means we create
            no parallel threads (i.e. we use a single thread), 0.5 means we create as many parallel threads as possible without
            increase the number of tokens we process by more than 50% (1.5 would mean 150%, etc.). Each thread process will use a
            different randomized LM prompt for test generation, so more threads will result in more diversity, but come at the cost
            of reading more prompt variations.

        prompt_builder : adatest.PromptBuilder
            A prompt builder to use when generating prompts for new tests. This object controls how the LM prompts
            are created when generating new tests.

        active_backend : "default", or a key name if adatest.backend is a dictionary
            Which backend from adatest.backend to use when generating new tests. This should always be set to "default" if
            adatest.backend is just a single backend and not a dictionary of backends.

        starting_path : str
            The path to start browsing the test tree from.

        embedding_model : sentencetransformer.EmbeddingModel
            A SentenceTransformer embedding model to use for semantic similarity.
        """

        # build the test tree browser
        return TestTreeBrowser(
            self,
            scorer=scorer,
            dataset=dataset,
            auto_save=auto_save,
            user=user,
            recompute_scores=recompute_scores,
            drop_inactive_score_columns=drop_inactive_score_columns,
            max_suggestions=max_suggestions,
            suggestion_thread_budget=suggestion_thread_budget,
            prompt_builder=prompt_builder,
            active_backend=active_backend,
            starting_path=starting_path,
            embedding_model=embedding_model,
            score_filter=score_filter,
            topic_model_scale=topic_model_scale
        )

    def __repr__(self):
        return self._tests.__repr__()

    def _repr_html_(self):
        return self._tests._repr_html_()

class TestTreeLocIndexer():
    def __init__(self, test_tree):
        self.test_tree = test_tree

    def __repr__(self):
        return "TestTreeLocIndexer is an intermediate object for operating on TestTrees. Slice this object further to yield useful results."

    def __getitem__(self, key):
        # If all columns haven't changed, it's still a valid test tree
        # If columns have been dropped, return a Pandas object
        
        subset = self.test_tree._tests.loc[key]
        if hasattr(subset, 'columns') and len(set(["type", "topic", "value1", "value2", "value3"]) - set(subset.columns)) == 0:
            test_tree_slice = TestTree(subset)
            test_tree_slice._tests_location = self.test_tree._tests_location
            return test_tree_slice
        else:
            return subset
    
    def __setitem__(self, key, value):
        self.test_tree._tests.loc[key] = value
    
class TestTreeILocIndexer():
    def __init__(self, test_tree):
        self.test_tree = test_tree

    def __repr__(self):
        return "TestTreeILocIndexer is an intermediate object for operating on TestTrees. Slice this object further to yield useful results."

    def __getitem__(self, key):
        # If all columns haven't changed, it's still a valid test tree
        # If columns have been dropped, return a Pandas object
        
        subset = self.test_tree._tests.iloc[key]
        if hasattr(subset, 'columns') and len(set(["type", "topic", "value1", "value2", "value3"]) - set(subset.columns)) == 0:
            test_tree_slice = TestTree(subset)
            test_tree_slice._tests_location = self.test_tree._tests_location
            return test_tree_slice
        else:
            return subset
    
    def __setitem__(self, key, value):
        self.test_tree._tests.iloc[key] = value


import time
from IPython.display import display, HTML
import openai
import numpy as np
import copy
import math
import sentence_transformers
import pandas as pd
import torch
import json
import re
import collections
from .comm import JupyterComm
import uuid
import pathlib
import copy
import re
import logging
import os
import io
import statistics
import checklist
import checklist.editor
from threading import Timer
from ._scorer import expand_template, clean_template, ClassifierScorer, GeneratorScorer, Scorer
from ._prompt_builder import PromptBuilder
import adatest

# from https://gist.github.com/walkermatt/2871026
def throttle(interval):
    """ Decorator that will postpone a functions
        execution so it does not run more than once per
        interval of time.
    """
    def decorator(fn):
        def throttled(*args, **kwargs):
            if not hasattr(throttled, "t") or not throttled.t.is_alive():
                def call_it():
                    fn(*args, **kwargs)
                throttled.t = Timer(interval, call_it)
                throttled.t.start()
        return throttled
    return decorator

log = logging.getLogger(__name__)

# import sys
# sys.stderr = open('/tmp/err.txt', 'w')

def file_log(*args):
    """ Used for logging when we don't have a stdout.

    This is used for debugging when we are being called from the javascript client. When we are
    called from the client we don't have a stdout attached to any cell in the notebook.

    Note to also catch errors you could do:
    import sys
    sys.stderr = open('err.txt', 'w')
    """
    #print(*args)
    f = open("log.txt", "a")  # append mode
    f.write(" ".join([str(msg) for msg in args])+"\n")
    f.flush()
    f.close()

cached_embedding_model = None

def is_subtopic(topic, candidate):
    # Returns true if candidate is a subtopic of topic
    return True if re.search(r'^%s(/|$)' % topic.replace('+', r'\+'), candidate) else False

special_outputs = [
    "{MAX}"
]

valid_comparators = [
    "should not be",
    "should be",
    "should be the same as for"
]
FILLIN_PREFIX = '/Fill-ins'


# model("this is english") => []
# output_sampling="topk(10)"
# output_sampling="topp(10)"
# output_sampling="max"
# output_sampling="temperature(0.9)"

class TestTreeBrowser():
    """ Used for browsing and expanding a test tree.
    """

    def __init__(self, test_tree, scorer, dataset, user, auto_save, recompute_scores, drop_inactive_score_columns,
                 max_suggestions, suggestion_thread_budget, prompt_builder, active_backend, starting_path,
                 embedding_model, score_filter, topic_model_scale):
        """ Initialize the TestTreeBrowser.
        
        See the __call__ method of TreeBrowser for parameter documentation.
        """

        self.test_tree = test_tree
        self.scorer = scorer
        self.dataset = dataset
        self.user = user
        self.auto_save = auto_save
        self.recompute_scores = recompute_scores
        self.drop_inactive_score_columns = drop_inactive_score_columns
        self.max_suggestions = max_suggestions
        self.suggestion_thread_budget = suggestion_thread_budget
        self.prompt_builder = prompt_builder
        self.active_backend = active_backend
        self.current_topic = starting_path
        self.embedding_model = embedding_model
        self.score_filter = score_filter
        self.topic_model_scale = topic_model_scale

        # get a reference to the active backend object
        if self.active_backend == "default":
            if isinstance(adatest.backend, dict):
                self._active_backend_obj = next(iter(adatest.backend.items()))[1]
            else:
                self._active_backend_obj = adatest.backend
        else:
            self._active_backend_obj = adatest.backend[self.active_backend]

        # if we are recomputing the scores then we erase all the old scores
        if recompute_scores is True:
            for c in self.test_tree.columns:
                if c.endswith("score"):
                    self.test_tree.drop(c, axis=1, inplace=True)

        # convert single scorer args to the multi-scorer format
        if callable(self.scorer):
            self.scorer = {"model": self.scorer}

        # note the score column of each scorer
        if isinstance(self.scorer, dict):
            self.score_columns = [k+" score" for k in self.scorer]
            for k in self.scorer:
                self.scorer[k] = Scorer(self.scorer[k])
        elif self.scorer is not None:
            self.score_columns = ["model score"]
            self.scorer = {"model": Scorer(self.scorer)}
        else:
            self.score_columns = []

        # find score columns that are not associated with a scorer
        for c in self.test_tree.columns:
            if c.endswith("score") and c not in self.score_columns:
                if drop_inactive_score_columns is True:
                    self.test_tree.drop(c, axis=1, inplace=True)
                else:
                    self.score_columns.append(c)

        # ensure that each scorer's score column is in the test tree dataframe
        for c in self.score_columns:
            if c not in self.test_tree.columns:
                self.test_tree[c] = [np.nan for _ in range(self.test_tree.shape[0])]

        # a unique identifier for this test set instance, used for UI connections
        self._id = uuid.uuid4().hex 

        # these are all temporary state
        self._embeddings = {} # Cached embedding vectors for each input/output pair (keyed on the pair ids)
        self._hidden_topics = {}
        self.comm = None

        # set the embedding model we use for similarity computations
        global cached_embedding_model
        if self.scorer is not None and embedding_model is None:
            if cached_embedding_model is None:
                cached_embedding_model = sentence_transformers.SentenceTransformer('stsb-roberta-base') # was large, not base in the past
            self.embedding_model = cached_embedding_model
        else:
            self.embedding_model = embedding_model

        # define our current mode, and set of supported modes
        self.mode = "tests"
        self.mode_options = [
            # "validity focused", # focus first on making valid in-topic tests, then secondarily on making those tests high scoring
            # "failure focused", # focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic
            "tests", # suggest new tests
            "topics" # suggest new subtopics
        ]
        
        self.dataset_preds, self.dataset_tests = None, None
        if dataset is not None:
            self.mode_options.append("dataset") # suggest new tests based on samples from the given dataset
            self.dataset_preds = self._load_dataset()
            self.dataset_tests = self._convert_dataset_to_tests(self.dataset_preds)
            self._compute_embeddings_and_scores(self.dataset_tests) # NOTE: Can be computationally intensive. Might re-compute scores.

        # make sure all the tests have scores (if we have a scorer)
        self._compute_embeddings_and_scores(self.test_tree)

        # save the current state of the test tree
        self._auto_save()

        # init a blank set of suggetions
        self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
        self._suggestions_error = "" # tracks if we failed to generate suggestions

    def _repr_html_(self, prefix="", environment="jupyter", websocket_server=None):
        """ Returns the HTML interface for this browser.

        Parameters
        ----------
        prefix : str
            The URL prefix this test tree browser is being served from.

        environment : str
            The environment this test tree browser is being served from (jupyter or web).
        """

        # spin up a JupyterComm object if we are called directly (which we assume is in a notebook)
        if self.comm is None and environment == "jupyter":
            self.comm = JupyterComm(f'adatest_interface_target_{self._id}', self.interface_event)

        # dump the client javascript to the interface
        file_path = pathlib.Path(__file__).parent.absolute()
        with open(file_path / ".." / "client" / "dist" / "main.js", encoding="utf-8") as f:
            js_data = f.read()
        interface_html = f"""
<div id="adatest_container_{self._id}" style="width: 100%; all: initial;"></div>
<script type='text/javascript'>
  {js_data};
  AdaTestReactDOM.render(
    AdaTestReact.createElement(AdaTest, {{
      interfaceId: "{self._id}", environment: "{environment}", startingTopic: "{self.current_topic}", prefix: "{prefix}",
      websocket_server: {"undefined" if websocket_server is None else '"'+websocket_server+'"'},\
    }}, null),
    document.getElementById('adatest_container_{self._id}')
  );
</script>
"""
        return interface_html

    def display(self):
        """ Manually display the HTML interface.
        """
        display(HTML(self._repr_html_()))

    def interface_event(self, msg):
        """ Handle interface events from the client.

        Parameters
        ----------
        msg : dict
            The event messages from the client. Each key in the dictionary is a separate message to either the row
            specified by the key or to whole browser object if the key is 'browser'.
        """

        log.debug(f"interface_event({msg})")

        # loop over each event message
        for k in msg:
            if k == "browser":
                action = msg[k].get("action", None)
                
                # rewdraw the entire interface
                if action == "redraw":
                    self._refresh_interface()
                
                # generate a new set of suggested tests/topics
                elif action == "generate_suggestions":
                    if self._active_backend_obj is None:
                        self._suggestions_error = "No AdaTest backend has been set!"
                    else:
                        # try:
                        self.suggestions = self._generate_suggestions(filter=msg[k].get("filter", ""))
                        self.suggestions.sort_values(self.score_columns[0], inplace=True, ascending=False, key=np.vectorize(score_max))
                        self._suggestions_error = ""
                        # except Exception as e:
                        #     log.debug(e)
                        #     self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
                        #     self._suggestions_error = True
                    self._refresh_interface()
                
                # change the current topic
                elif action == "change_topic":
                    self.current_topic = msg[k]["topic"]
                    self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)

                    # see if we have only topics are direct children, if so, we suggest topics, otherwise we suggest tests
                    has_direct_tests = False
                    has_known_subtopics = False
                    for k, test in self.test_tree.iterrows():
                        if test["topic"] == self.current_topic:
                            if test["type"] == "test":
                                has_direct_tests = True
                        elif is_subtopic(self.current_topic, test["topic"]):
                            has_known_subtopics = True
                    if not has_direct_tests and has_known_subtopics:
                        self.mode = "topics"
                    else:
                        self.mode = "tests"

                    self._refresh_interface()
                
                # clear the current set of suggestions
                elif action == "clear_suggestions":
                    self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
                    self._refresh_interface()

                # add a new empty subtopic to the current topic
                elif action == "add_new_topic":
                    self.test_tree.loc[uuid.uuid4().hex] = {
                        "topic": self.current_topic + "/New topic",
                        "type": "topic_marker",
                        "value1": "",
                        "value2": "",
                        "value3": "",
                        "author": self.user,
                        "description": ""
                    }
                    self._compute_embeddings_and_scores(self.test_tree)
                    self._auto_save()
                    self._refresh_interface()
                
                # add a new empty test to the current topic
                elif action == "add_new_test":
                    
                    # find the common values and type in this subtopic
                    types = []
                    value2s = []
                    value3s = []
                    for k, test in self.test_tree.iterrows():
                        if is_subtopic(self.current_topic, test.topic) and test.type != "topic_marker":
                            types.append(test.type)
                            if test.value2 != "":
                                value2s.append(test.value2)
                            if test.value3 != "":
                                value3s.append(test.value3)
                    if len(types) == 0:
                        types = ["{} should not output {}"]
                    if len(value2s) == 0:
                        value2s = [""]
                    if len(value3s) == 0:
                        value3s = [""]

                    # add the new test row
                    row = {
                        "topic": self.current_topic,
                        "type": safe_mode(types),
                        "value1": "New test", # The special value "New test" causes the interface to auto-select the text
                        "value2": safe_mode(value2s),
                        "value3": safe_mode(value3s),
                        "author": self.user,
                        "description": ""
                    }
                    for c in self.score_columns:
                        row[c] = np.nan
                        row[c[:-6] + " value1 outputs"] = "{}"
                        row[c[:-6] + " value2 outputs"] = "{}"
                        row[c[:-6] + " value3 outputs"] = "{}"
                    self.test_tree.loc[uuid.uuid4().hex] = row

                    self._compute_embeddings_and_scores(self.test_tree)
                    self._auto_save()
                    self._refresh_interface()

                # change which scorer/model is used for sorting tests
                elif action == "set_first_model":
                    name = msg[k]["model"]

                    # move to front of score columns in data frames
                    for df in [self.suggestions, self.test_tree]:
                        pos = len(df.columns) - len(self.score_columns)
                        tmp = df[name]
                        df.drop(labels=[name], axis=1, inplace=True)
                        df.insert(pos, name, tmp)

                    # update score columns list
                    self.score_columns.remove(name)
                    self.score_columns.insert(0, name)

                    self._auto_save()
                    self._refresh_interface()

                # change which backend is active
                elif action is None and "active_backend" in msg[k]:
                    self.active_backend = msg[k]["active_backend"]
                    self._active_backend_obj = adatest.backend[self.active_backend]

                # change which backend is active
                elif action is None and "mode" in msg[k]:
                    self.mode = msg[k]["mode"]

            # if we are just updating a single row in suggestions or tests then we only recompute the scores
            elif "topic" not in msg[k]:
                df = None
                if k in self.suggestions.index:
                    df = self.suggestions
                elif k in self.test_tree.index:
                    df = self.test_tree
                if df is not None:
                    sendback_data = {"author": self.user}
                    
                    # convert template expansions into a standard value update
                    if msg[k].get("action", "") == "template_expand":
                        template_value = self.templatize(df.loc[k, msg[k]["value"]])
                        msg[k] = {msg[k]["value"]: template_value}
                        sendback_data[msg[k]["value"]] = template_value

                    # update the row and recompute scores
                    for k2 in msg[k]:
                        df.loc[k, k2] = msg[k][k2]
                    df.loc[k, self.score_columns] = None
                    if k in self._embeddings:
                        del self._embeddings[k]
                    self._compute_embeddings_and_scores(df)
                    self._auto_save()

                    # send just the data that changed back to the frontend
                    sendback_data["scores"] = {c: [[k, v] for v in score_parts(df.loc[k, c])] for c in self.score_columns}
                    for value in ["value1", "value2", "value3"]:
                        outputs = {c: [[k, json.loads(df.loc[k].get(c[:-6] + " "+value+" outputs", "{}"))]] for c in self.score_columns}
                        sendback_data[value+"_outputs"] = outputs
                    sendback_data.update(self.test_display_parts(df.loc[k]))
                    self.comm.send({k: sendback_data})

            # if we are just changing the topic
            elif "topic" in msg[k] and len(msg[k]) == 1:

                # move a test that is in the suggestions list
                if k in self.suggestions.index:
                    self.suggestions.loc[k, "topic"] = msg[k]["topic"]
                    self.suggestions.loc[k, "author"] = self.user
                    self.test_tree.loc[k] = self.suggestions.loc[k]
                    self.suggestions.drop(k, inplace=True)
                
                # move a test that is in the test tree
                elif k in self.test_tree.index:
                    if msg[k]["topic"] == "_DELETE_": # this means delete the test
                        self.test_tree.drop(k, inplace=True)
                    elif msg[k]["topic"] == "suggestion": # this means move the test back to the suggestions list
                        self.test_tree.loc[k, "topic"] = msg[k]["topic"]
                        self.suggestions.loc[k] = self.test_tree.loc[k]
                        self.test_tree.drop(k, inplace=True)
                    else:
                        self.test_tree.loc[k, "topic"] = msg[k]["topic"]
                        self.test_tree.loc[k, "author"] = self.user
                
                # move a whole topic around
                else:
                    for id, test in self.test_tree.iterrows():
                        if is_subtopic(k, test.topic):
                            if msg[k]["topic"] == "suggestion":
                                self.test_tree.loc[id, "topic"] = msg[k]["topic"]
                                self.suggestions.loc[id] = self.test_tree.loc[id]
                                self.test_tree.drop(id, inplace=True)
                            else:
                                self.test_tree.loc[id, "topic"] = msg[k]["topic"] + test.topic[len(k):]

                            if test.type == 'topic_marker' and test.topic == k:
                                if id in self._embeddings:
                                    del self._embeddings[id]
                    # Move topic out of suggestions into tests
                    for id, test in self.suggestions.iterrows():
                        if is_subtopic(k, test.topic):
                            if msg[k]["topic"] != "suggestion":
                                self.suggestions.loc[id, "topic"] = msg[k]["topic"]
                                self.test_tree.loc[id] = self.suggestions.loc[id]
                                self.suggestions.drop(id, inplace=True)

                # Recompute any missing embeddings to handle any changes
                self._compute_embeddings_and_scores(self.test_tree)
                self._compute_embeddings_and_scores(self.suggestions)
                self._auto_save()
                self._refresh_interface()

            else:
                log.debug(f"Unable to parse the interface message: {msg[k]}")

    def _refresh_interface(self):
        """ Send our entire current state to the frontend interface.
        """

        # get the children of the current topic
        data = {}

        def create_children(data, tests):
            children = []
            
            # add tests and topics to the data lookup structure
            for k, test in tests.iterrows():
                if is_subtopic(self.current_topic, test.topic):
                    
                    # add a topic
                    if test.type == "topic_marker":
                        if is_subtopic(self.current_topic, test.topic) and test.topic != self.current_topic:
                            name = test.topic[len(self.current_topic)+1:]
                            if "/" not in name: # only add direct children
                                data[test.topic] = {
                                    "type": test.type,
                                    "author": test.author,
                                    "description": "",
                                    "scores": {c: [] for c in self.score_columns},
                                    "topic_marker_id": k,
                                    "topic_name": name,
                                    "editing": test.topic.endswith("/New topic")
                                }
                                children.append(test.topic)
                    
                    # add a test
                    else:
                        data[k] = {
                            "type": test.type,
                            "author": test.author,
                            "description": test.description,
                            "scores": {c: [[k, v] for v in score_parts(test[c])] for c in self.score_columns},
                            "editing": test.value1 == "New test"
                        }
                        for value in ["value1", "value2", "value3"]:
                            data[k][value] = test[value]
                            data[k][value+"_outputs"] = {c: [[k, safe_json_load(test.get(c[:-6] + " "+value+" outputs", "{}"))]] for c in self.score_columns}
                        data[k].update(self.test_display_parts(test))
                        if test.topic == self.current_topic:
                            children.append(k)
            
            # fill in the scores for the child topics
            for k, test in tests.iterrows():
                if is_subtopic(self.current_topic, test.topic) and test.topic != self.current_topic:
                    child_topic = test.topic[len(self.current_topic):].split("/", 2)[1]
                    scores = data[self.current_topic+"/"+child_topic]["scores"]
                    for c in self.score_columns:
                        scores[c].extend([[k, v] for v in score_parts(test[c])])

            # sort by score and always put new topics first
            def sort_key(id):
                total = 0
                count = 0
                for s in data[id]["scores"][self.score_columns[0]]:
                    val = score_max(s[1], nan_val=np.nan)
                    if not np.isnan(val):
                        total += val
                        count += 1
                if count == 0:
                    return 1e3
                else:
                    return -total / count
            sorted_children = sorted(children, key=sort_key)
            sorted_children = sorted(sorted_children, key=lambda id: 0 if id.endswith("/New topic") or data[id].get("value1", "") == "New test" else 1)

            return sorted_children
        
        # get the children of the current topic
        children = create_children(data, self.test_tree)
        suggestions_children = create_children(data, self.suggestions)

        # TODO: This is a complete hack to hide lower scoring suggestions when we are likely already in the exploit phase
        # this is just for users who don't know when to stop scrolling down...
        # SML: I expect we can delete this at some point?
        if self.score_filter == "auto":
            if len(children) < 10:
                score_filter = -1e12
            else:
                children_scores = sorted([np.max([score_max(x[1]) for x in data[key]['scores'][self.score_columns[0]]]) for key in children])
                suggestions_children_scores = sorted([np.max([score_max(x[1]) for x in data[key]['scores'][self.score_columns[0]]]) for key in suggestions_children])
                score_filter = children_scores[-5] - (children_scores[-1] - children_scores[-5]) * 0.2
                if len(suggestions_children_scores) > 0:
                    score_filter = min(score_filter, np.nanmax(suggestions_children_scores) - 1e-2)
        else:
            score_filter = self.score_filter

        # compile the global browser state for the frontend
        data["browser"] = {
            "suggestions": suggestions_children,
            "tests": children,
            "topic": self.current_topic,
            "topic_description": data[self.current_topic]["description"] if self.current_topic in data else "",
            "topic_marker_id": data[self.current_topic]["topic_marker_id"] if self.current_topic in data else uuid.uuid4().hex,
            "score_filter": score_filter,
            "disable_suggestions": False,
            "read_only": False,
            "score_columns": self.score_columns,
            "suggestions_error": self._suggestions_error,
            "backend_options": adatest.backend.keys() if isinstance(adatest.backend, dict) else [self.active_backend],
            "active_backend": self.active_backend,
            "mode": self.mode,
            "mode_options": self.mode_options,
            "test_types": self.scorer[self.score_columns[0][:-6]].supported_test_types,
            "test_type_parts": {t: split_test_type(t) for t in self.scorer[self.score_columns[0][:-6]].supported_test_types},
        }

        self.comm.send(data)

    def _generate_suggestions(self, filter):
        """ Generate suggestions for the current topic.

        Parameters
        ----------
        filter : str
            The filter to apply to the tests while generating suggestions.
        """

        if self.mode == "dataset":
            current_tests = self.test_tree[(self.test_tree['topic'] == self.current_topic) & (self.test_tree['type'] != 'topic_marker')]
            if len(current_tests) == 0: # If empty or all topic_markers
                # Just return largest error samples here
                error_indices = self.dataset_preds.sort_values(by='largest_error_proba', ascending=False).head(n=self.max_suggestions).index
                output = self.dataset_tests.loc[error_indices].copy()
                output['topic'] = self.current_topic
                return output

            # Otherwise find topics closest to the current tests in topic
            self._compute_embeddings_and_scores(current_tests) # Make sure the current tests have embeddings calculated
            topic_embeddings = torch.vstack([torch.tensor(self._embeddings[k]) for k in current_tests.index])
            data_embeddings = torch.vstack([torch.tensor(self._embeddings[k]) for k in self.dataset_tests.index])
            
            method = 'distance_to_avg' # TODO: Pick one, just leaving both for experimentation
            if method == 'avg_distance':
                dist = sentence_transformers.util.pytorch_cos_sim(topic_embeddings, data_embeddings)
                closest_indices = torch.topk(dist.mean(axis=0), k=self.max_suggestions).indices
                
            elif method == 'distance_to_avg':
                avg_topic_embedding = topic_embeddings.mean(axis=0)

                distance = sentence_transformers.util.pytorch_cos_sim(avg_topic_embedding, data_embeddings)
                closest_indices = torch.topk(distance, k=self.max_suggestions).indices

            output = self.dataset_tests.iloc[np.array(closest_indices).squeeze()].copy()
            output['topic'] = self.current_topic
            return output

        #--Backend-driven suggestions--

        # save a lookup we can use to detect duplicate tests
        test_map = {}
        for _, test in self.test_tree.iterrows():
            if test.type == "topic_marker":
                parts = test.topic.rsplit("/", 1)
                if len(parts) == 2:
                    value1 = parts[1]
                else:
                    value1 = ""
            else:
                value1 = test.value1
            test_map[test.type + " " + value1 + " " +  test.value2 + " " +  test.value3] = True

        # see if we have a finite set of valid outputs
        # valid_outputs = getattr(self.scorer, "output_names", None)
        # if valid_outputs is not None and value2_filter is not None:
        #     valid_outputs = [s for s in valid_outputs if re.search(value2_filter, s) is not None]

        # # see if we have only topics are direct children, if so, we suggest topics
        # has_direct_tests = False
        # has_known_subtopics = False
        # for k, test in self.test_tree.iterrows():
        #     if test["topic"] == self.current_topic:
        #         if test["type"] == "test":
        #             has_direct_tests = True
        #     elif is_subtopic(self.current_topic, test["topic"]):
        #         has_known_subtopics = True
        # suggest_topics = not has_direct_tests and has_known_subtopics
        # suggest_topics = self.mode == "topics"
        # zero_shot_tests = not has_direct_tests and not has_known_subtopics
        
        
        
        # {} should output {}
        # {} should not output {}
        # {} should have the same output as {}
        # {} should not be more {} than {}
        # {} should be invertable. [] is the output.
        # {} should not output text containing {}
        # {} should output toxic text. [] is the output.
        # {} should not output toxic text. [] is the output.
        
        # {}'s output should not contain {}
        


        
        # validity focused (focus first on making valid in-topic tests, then secondarily on making those tests high scoring)
        # failure focused (focus on making high scoring (failing) tests, then secondarily on making those tests valid and in-topic)
        # topics (suggest new sub-topics)
        # file_name dataset (suggest tests based on samples from the provided dataset)
        


        # compute the maximum number of suggestion threads we can use given our suggestion_thread_budget
        p = self.prompt_builder.prompt_size
        budget = 1 + self.suggestion_thread_budget
        suggestion_threads =  max(1, int(np.floor(budget * (p/(p+1) + 1/(p+1) * self.max_suggestions) - 1/(p+1) * self.max_suggestions) / (p/(p+1))))
        
        # generate the prompts for the backend
        test_type, prompts = self.prompt_builder(
            test_tree=self.test_tree,
            topic=self.current_topic,
            score_column=self.score_columns[0],
            repetitions=suggestion_threads,
            filter=filter,
            suggest_topics=self.mode == "topics",
            embeddings=self._embeddings
        )

        # generate the suggestions
        proposals = self._active_backend_obj(prompts, self.current_topic, test_type, self.scorer, num_samples=self.max_suggestions // len(prompts))

        log.debug(f"proposals = {proposals}")

        # filter out suggestions that are duplicates before we score them
        suggestions = []
        test_map_tmp = copy.copy(test_map)
        for value1, value2, value3 in proposals:
            if self.mode == "topics" and ("/" in value1 or "\n" in value1):
                value1 = value1.replace("/", " or ").replace("\n", " ") # topics can't have newlines or slashes in their names
                value1 = value1.replace("  ", " ").strip() # kill any double spaces we may have introduced
            str_val = test_type + " " + value1 + " " + value2 + " " + value3
            if str_val not in test_map_tmp:
                s = {
                    "type": test_type,
                    "topic": self.current_topic + ("/"+value1 if self.mode == "topics" else ""),
                    "value1": "" if self.mode == "topics" else value1,
                    "value2": value2,
                    "value3": value3,
                    "author": self._active_backend_obj.__class__.__name__ + " backend",
                    "description": ""
                }
                for c in self.score_columns:
                    s[c] = np.nan
                suggestions.append(s)
                if str_val is not None:
                    test_map_tmp[str_val] = True

        suggestions = pd.DataFrame(suggestions, index=[uuid.uuid4().hex for _ in range(len(suggestions))], columns=self.test_tree.columns)
        self._compute_embeddings_and_scores(suggestions)
        if self.mode != "topics":
            suggestions = suggestions.dropna(subset=[self.score_columns[0]])

        # When we have outputs filled in by the scorer we might have more duplicates we need to remove
        duplicates = []
        for k,row in suggestions.iterrows():
            str_val = row.topic + " " + test_type + " " + row.value1 + " " +  row.value2 + " " +  row.value3
            if str_val in test_map:
                duplicates.append(k)
            test_map[str_val] = True
        suggestions = suggestions.drop(duplicates)

        if self.topic_model_scale != 0:
            self._add_topic_model_score(suggestions, topic_model_scale=self.topic_model_scale)
        return suggestions

    def _add_topic_model_score(self, df, topic_model_scale):
        documents = []
        for k,s in df.iterrows():
            max_output = -10e8
            max_output_name = None
            for k,v in json.loads(s["score value1 outputs"]).items():
                if v > max_output:
                    max_output = v
                    max_output_name = k
            documents.append(f'"{s["value1"]}" > "{max_output_name}"')

        query = self._make_prompt(
            self.current_topic,
            prompt_size=20,
            include_value2=True
        )["prompt"]

        r = openai.Engine("davinci-instruct-beta").search(
            documents=documents,
            query=query
        )

        sim_scores = np.array([v["score"] for v in r["data"]])
        sim_scores -= np.mean(sim_scores)
        sim_scores /= np.std(sim_scores)

        for i, (k, row) in enumerate(df.iterrows()):
            row["score"] = float(row["score"]) + topic_model_scale * sim_scores[i]

    def _compute_embeddings_and_scores(self, tests, recompute=False):
        log.debug(f"compute_embeddings_and_scores(tests=<DataFrame shape={tests.shape}>, recompute={recompute})")

        if self.scorer is not None:
            self._compute_scores(tests, recompute=recompute)

        # model outputs and embeddings
        if self.embedding_model is not None:
            new_embedding_ids = [k for k in tests.index if k not in self._embeddings]
            if len(new_embedding_ids) > 0:
                value1s = []
                value2s = []
                value3s = []
                for k in new_embedding_ids:
                    if tests.loc[k, "type"] == "topic_marker":
                        parts = tests.loc[k, "topic"].rsplit("/", 1)
                        value1s.append(parts[1] if len(parts) == 2 else "")
                        value2s.append("")
                        value3s.append("")
                    else:
                        value1s.append(str(tests.loc[k, "value1"]))
                        value2s.append(str(tests.loc[k, "value2"]))
                        value3s.append(str(tests.loc[k, "value3"]))
                new_value1_embeddings = self.embedding_model.encode(value1s, convert_to_tensor=True, show_progress_bar=False).cpu()
                new_value2_embeddings = self.embedding_model.encode(value2s, convert_to_tensor=True, show_progress_bar=False).cpu()
                new_value3_embeddings = self.embedding_model.encode(value3s, convert_to_tensor=True, show_progress_bar=False).cpu()
                for i,k in enumerate(new_embedding_ids):
                    self._embeddings[k] = np.hstack([new_value1_embeddings[i], new_value2_embeddings[i], new_value3_embeddings[i]])

    def _compute_scores(self, tests, recompute):
        """ Use the scorer(s) to fill in scores in the passed TestTree.

        Parameters
        ----------
        tests : TestTree
            The TestTree to fill in missing scores for.

        recompute : bool
            If True, recompute all scores. If False, only recompute scores that are missing.
        """
        
        log.debug(f"_compute_scores(tests=<TestTree shape={tests.shape}>, recompute={recompute})")

        # see which rows need scores computed
        if recompute or len(self.score_columns) == 0:
            new_score_mask = np.ones(tests.shape[0], dtype=np.bool)
        else:
            new_score_mask = np.array(tests[self.score_columns[0]].isnull()) | np.array(tests[self.score_columns[0]] == "")
        new_score_mask = new_score_mask & np.array(tests["type"] != "topic_marker", dtype=np.bool)

        if new_score_mask.sum() > 0:
            scores = {}
            tests_to_score = tests.loc[new_score_mask, ["type", "value1", "value2", "value3"]]
            
            # call the scorers
            blank_outputs = [{} for _ in range(tests_to_score.shape[0])]
            for k in self.scorer:
                scorer_output = self.scorer[k](tests_to_score)
                scores[k+" score"] = ["|".join(str(vv) for vv in v) for v in scorer_output["scores"]]
                scores[k+" value1 outputs"] = scorer_output.get("value1_outputs", blank_outputs)
                scores[k+" value2 outputs"] = scorer_output.get("value2_outputs", blank_outputs)
                scores[k+" value3 outputs"] = scorer_output.get("value3_outputs", blank_outputs)

            # copy the scores into the TestTree
            for k in scores:
                for i, j in enumerate(np.where(new_score_mask)[0]):
                    tests.loc[tests.index[j], k] = json.dumps(scores[k][i]) if isinstance(scores[k][i], dict) else scores[k][i]

            # copy outputs that may have been generated by the scorers over to the passed test tree
            for k in tests.index[new_score_mask]:
                tests.loc[k, "value1"] = tests_to_score.loc[k, "value1"]
                tests.loc[k, "value2"] = tests_to_score.loc[k, "value2"]
                tests.loc[k, "value3"] = tests_to_score.loc[k, "value3"]

    def _load_dataset(self, time_budget=30, min_samples=100):
        '''Evaluate model on dataset and capture useful information.'''
        # TODO: Generalize to more dataset formats
        if self.dataset is None:
            return None
        
        model = self.scorer['model'].model

        # Unpack dataset object
        X, y = self.dataset[0], self.dataset[1]
        output_names = self.scorer['model'].output_names
        
        unknown_labels = set(y) - set(output_names)
        assert len(unknown_labels) == 0, f"Unknown labels found: {unknown_labels}. \
        Please update the label vector or output names property."

        # Time how long inference takes on a single sample
        try:
            start = time.time()
            _ = model(X[0:1])
            end = time.time()
        except Exception as e: # TODO: Improve this message
            raise ValueError(f"Training data cannot be evaluated by model. Error recieved: {e}.")


        # Ensure min_samples <= n_samples <= len(data) and computes in {time_budget} seconds
        n_samples = int(min(max(time_budget // (end - start), min_samples), len(X)))

        if n_samples < len(X):
            print(f"Only using {n_samples} samples to meet time budget of {time_budget} seconds.")
            # TODO: unify input types
            sample_indices = np.random.choice(np.arange(len(X)), n_samples, replace=False)
            X = [X[sample] for sample in sample_indices]
            y = [y[sample] for sample in sample_indices]

        # Build output frame
        df = pd.DataFrame(columns=['sample', 'label', 'label_proba', \
                                            'pred', 'pred_proba', 'largest_error', 'largest_error_proba'])
        df['sample'] = X
        df['label'] = y

        # model's current prediction
        raw_model_output = model(X)
        pred_indices = np.argsort(raw_model_output, axis=1)
        
        df['pred_proba'] = raw_model_output[range(len(pred_indices)), pred_indices[:, -1]]
        df['pred'] = [output_names[i] for i in pred_indices[:, -1]]

        label_lookup = {output:index for index, output in enumerate(output_names)}
        label_indices = [label_lookup[label] for label in y]
        df['label_proba'] = raw_model_output[range(len(label_indices)), label_indices]
        

        correct_predictions = df['pred'] == df['label']
        mispredictions = ~correct_predictions
        
        # For mispredicted samples, the largest error is the current prediction.
        df.loc[mispredictions, 'largest_error'] = df.loc[mispredictions, 'pred']
        df.loc[mispredictions, 'largest_error_proba'] = df.loc[mispredictions, 'pred_proba']
        
        # For correct samples, we use the 2nd highest class as the largest error.
        largest_errors = pred_indices[correct_predictions][:, -2]
        df.loc[correct_predictions, 'largest_error'] = [output_names[i] for i in largest_errors]
        df.loc[correct_predictions, 'largest_error_proba'] = raw_model_output[range(len(largest_errors)), largest_errors]

        df.index = [uuid.uuid4().hex for _ in range(len(df))]
        return df

    def _convert_dataset_to_tests(self, dataset_frame): # TODO: Consider removing from class?
        '''Converts a loaded dataset into test formats.'''
        
        column_names = ['topic', 'type' , 'value1', 'value2', 'value3', 'author', 'description', \
        'model value1 outputs', 'model value2 outputs', 'model value3 outputs', 'model score']

        test_frame = pd.DataFrame(columns=column_names)

        # All tests currently formatted as not predicting the largest error.
        test_frame['value1'] = dataset_frame['sample']
        test_frame['type'] = "{} should not output {}"
        test_frame['value2'] = dataset_frame['largest_error']

        # Constants
        test_frame['topic'] = ''
        test_frame['author'] = "dataset"
        test_frame['description'] = ''

        test_frame.index = dataset_frame.index
        
        return test_frame # TODO: Cast this as a formal TestTree instead of dataframe

    def test_display_parts(self, test):
        if test.type == "{} should not be completed to become {}":
            return {
                "d_text1a": '"',
                "d_value1": "{}",
                "d_text1b": safe_json_load(test.get(self.score_columns[0][:-6] + " value1 outputs", "{}")).get("string", "") + '"',
                "d_text2a": '"',
                "d_value2": "{}",
                "d_text2b": '"',
                "d_text3a": '',
                "d_value3": "",
                "d_text3b": ''
            }
        else: # this is the default two-value test format that only varies in the select value
            return {
                "d_text1a": '"',
                "d_value1": "{}",
                "d_text1b": '"',
                "d_text2a": '"',
                "d_value2": "{}",
                "d_text2b": '"',
                "d_text3a": '',
                "d_value3": "",
                "d_text3b": ''
            }

    def templatize(self, s):
        prompt = """INPUT: "Where are regular people on Twitter"
    OUTPUT: "Where are {regular|normal|sane|typical} people on {Twitter|Facebook|Reddit|Instagram}"
    ###
    INPUT: "Anyone who says this food tastes horrible is out of their mind"
    OUTPUT: "{Anyone|Someone|He|She} who says this food tastes {horrible|terrible|rotten} is out of their mind"
    ###
    INPUT: "great"
    OUTPUT: "{great|excellent|wonderful|superb|delightful}"
    ###
    INPUT: "If you haven't come here before, you probably live under a rock"
    OUTPUT: "If you haven't come here {before|in the past|before now|yet}, you probably {live under a rock|never get out|are a hermit|are isolated}"
    ###
    INPUT: "Only crazy people would say they had a lousy time"
    OUTPUT: "Only {crazy people|insane people|people with no sense|those out of their mind} would say they had a {lousy|terrible|bad|miserable} time"
    ###
    INPUT: "If I didn't come here again, I would be very happy for the rest of my life"
    OUTPUT: "If I didn't come {here|hereabouts|around here} {again|once more|all over again}, I would be very {happy|glad|pleased|elated} for the rest of my life"
    ###
    INPUT: "I don't know what James was talking about when they said they loved the food."
    OUTPUT: "I don't know what {James|John|Robert|Steve|Bob} was talking about when they {said they|stated that|claimed that|mentioned that} they {loved|liked|adored|appreciated} the food."
    ###
    INPUT: "new_input_value"
    OUTPUT: \""""
        prompt = prompt.replace("new_input_value",  s)
        response = openai.Completion.create(
            engine=self.engine, prompt=prompt, max_tokens=300,
            temperature=0.7, n=4, stop="\""
        )

        lines = [choice["text"] for choice in response["choices"]]
        options = []
        for line in lines:
            line = clean_template(line)
            valid = False
            for option in expand_template(line):
                if option == s:
                    valid = True
                    break
            if valid:
                options.append((-len(line), line))
        options.sort()
        log.debug(f"options = {options}")
        return options[0][1]
    
    def _auto_save(self):
        """ Save the current state of the model if we are auto saving.
        """
        if self.auto_save:
            self.test_tree.to_csv()

def score_max(s, nan_val=-1e3):
    if s == "":
        return nan_val
    elif isinstance(s, str):
        return np.max([convert_float(v) for v in s.split("|")])
    elif np.isnan(s):
        return nan_val
    else:
        return np.max(s)

def score_parts(s):
    if isinstance(s, str):
        return [convert_float(v) for v in s.split("|")]
    else:
        return [s]

def convert_float(s):
    try:
        f = float(s)
    except ValueError:
        f = np.nan
    return f

def safe_json_load(input):
    if isinstance(input, float): # catch NaN's
        return {}
    else:
        return json.loads(input)

def split_test_type(test_type):
    part_names = ["text1", "value1", "text2", "value2", "text3", "value3", "text4"]
    parts = re.split(r"(\{\}|\[\])", test_type)
    part_values = ["" for _ in range(7)]
    for i, part in enumerate(parts):
        part_values[i] = part
    return {name: value for name,value in zip(part_names, part_values)}



def safe_mode(l):
    """ This just silences the error from a double mode from python <= 3.7.
    """
    try:
        return statistics.mode(l)
    except:
        return l[0]


import os
import re
import codecs
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

with open('requirements.txt') as f:
    install_requires = [line.strip() for line in f]
    

setup(
    name='adatest',
    version=find_version("adatest", "__init__.py"),
    url='https://github.com/microsoft/adatest.git',
    author='Scott Lundberg and Marco Tulio Ribeiro',
    author_email='scott.lundberg@microsoft.com',
    description='Adaptively test and debug any natural language machine learning model.',
    packages=find_packages(exclude=['user_studies', 'notebooks', 'client']),
    install_requires=install_requires
)



import sklearn
import numpy as np
from sklearn import multioutput
from sklearn import preprocessing
from sklearn.linear_model import RidgeClassifierCV

class ChainTopicModel:
    def __init__(self, model=None):
        if model is None:
            self.base_model = RidgeClassifierCV()
        else:
            self.base_model = model
    def fit(self, X, y):
        topics = y
        max_levels = max([len(x.split('>')) for x in topics])
        self.model = sklearn.multioutput.ClassifierChain(self.base_model, order=list(range(max_levels)))
        y = [list(map(str.strip, x.split('>'))) for x in topics]
        y = np.array([x + ['-'] * (max_levels - len(x)) for x in y])
        self.encoders = [preprocessing.LabelEncoder() for _ in range(max_levels)]
        self.possible_topics = set()
        for x in topics:
            self.possible_topics.add(x)
            a = x.split(' > ')
            for i in range(1, len(a)):
                self.possible_topics.add(' > '.join(a[:i]))

        self.classes_ = list(self.possible_topics)
        new_y = np.zeros(y.shape)
        for i in range(y.shape[1]):
            self.encoders[i].fit(y[:, i])
            new_y[:, i] = self.encoders[i].transform(y[:, i])
        self.model.fit(X, new_y)
    def predict(self, X):
        y = self.model.predict(X)
        ret = []
        for i in range(y.shape[1]):
            ret.append(self.encoders[i].inverse_transform(y[:, i].astype(int)))
        y = np.array(ret).T
        ret = []
        for x in y:
            x = [z for z in x if z != '-']
            a = ' > '.join(x)
            while a not in self.possible_topics:
                x = x[:-1]
                a = ' > '.join(x)
            ret.append(a)
        return np.array(ret)

    def predict_proba(self, X):
        # This is just a fake function for now, puts 1 in the predicted class and 0 elsewhere
        y = self.predict(X)
        ret = np.zeros((len(X), len(self.classes_)))
        for i, r in enumerate(y):
            ret[i, self.classes_.index(r)] = 1
        return ret

class StandardTopicModel:
    def __init__(self, threshold=0.5):
        self.model= sklearn.linear_model.RidgeClassifierCV()
        self.threshold=threshold
        # add the missing predict_proba method to RidgeClassifierCV
        def predict_proba(self, X):
            if len(self.classes_) == 1:
                return np.ones((len(X), 1))
            d = self.decision_function(X)
            if len(self.classes_) == 2:
                probs = np.exp(d) / (np.exp(d) + np.exp(-d))
                return np.array([1 - probs, probs]).T
            probs = np.exp(d).T / np.sum(np.exp(d), axis=1)
            return probs.T
        self.model.predict_proba = predict_proba.__get__(self.model, self.model.__class__)
    def fit(self, X, y):
        self.model.fit(X, y)
    def predict_proba(self, X):
        return self.model.predict_proba(X)
    def predict(self, X):
        if self.threshold is None:
            return self.model.predict(X)
        pps = self.model.predict_proba(X)
        zero_index = list(self.model.classes_).index('Not problematic')
        ret = []
        for p in pps:
            if p[zero_index] >= self.threshold:
                ret.append(self.model.classes_[zero_index])
                continue
            else:
                best = np.argsort(p)
                if best[-1] == zero_index:
                    best = best[:-1]
                ret.append(self.model.classes_[best[-1]])
        return np.array(ret)
        # return self.model.predict(X)



from ._test_tree import TestTree
from ._test_tree_browser import TestTreeBrowser
from ._scorer import Scorer, DummyScorer, ClassifierScorer, GeneratorScorer
from ._server import serve
from ._model import Model
from ._topic_model import ChainTopicModel, StandardTopicModel
from . import backends

__version__ = '0.0.3'

backend = None


import numpy as np
import shap
import transformers
from .utils import isinstance_ipython


class Model():
    """ This wraps models used in AdaTest so that have a consistent interface.

    This should eveutually just be the Model class from SHAP, but we keep a simple version here for now
    so we can easily update it during initial development.
    """

    def __new__(cls, model, *args, **kwargs):
        """ If we are wrapping a model that is already a Model, we just return it.
        """
        if isinstance_ipython(model, Model) or isinstance_ipython(model, shap.models.Model):
            return model
        else:
            return super().__new__(cls)
    
    def __init__(self, model, output_names=None, **kwargs):
        """ Build a new model by wrapping the given model object.

        Parameters
        ----------
        model : object
            The model to wrap. This can be a plain python function that accepts a list of strings and returns either
            a vector of probabilities or another string. It can also be a transformers pipeline object (we try to wrap
            common model types transparently).

        output_names : list of str, optional
            The names of the outputs of the model. If not given, we try to infer them from the model.
        """

        # finish early if we are wrapping an object that is already a Model
        if isinstance_ipython(model, Model) or isinstance_ipython(model, shap.models.Model):
            if output_names is not None:
                self.output_names = output_names
            assert len(kwargs) == 0
            return

        # get outputs names from the model if it has them and we don't
        if output_names is None and hasattr(model, "output_names"):
            output_names = model.output_names

        # If we are in the base class we check to see if we should rebuild the model as a specialized subclass
        if self.__class__ is Model:
            
            # wrap transformer pipeline objects for convenience
            if isinstance_ipython(model, transformers.pipelines.text_classification.TextClassificationPipeline):
                self.__class__ = shap.models.TransformersPipeline
                shap.models.TransformersPipeline.__init__(self, model, **kwargs)
                if output_names is not None: # Override output names if user supplied
                    self.output_names = output_names

            elif isinstance_ipython(model, transformers.pipelines.text_generation.TextGenerationPipeline):
                self.__class__ = TransformersTextGenerationPipeline
                TransformersTextGenerationPipeline.__init__(self, model, **kwargs)
            
            else:
                self.inner_model = model
                self.output_names = output_names

    def __call__(self, *args):
        return np.array(self.inner_model(*args))

    
class TransformersTextGenerationPipeline():
    """ This wraps the transformer text generation pipeline object to match the Model API.

    TOTO: move this to SHAP.
    """
    def __init__(self, pipeline):
        self._inner_model = pipeline
        self.output_names = None

    def __call__(self, strings):
        inner_out = self._inner_model(strings)
        out = []
        for s, data in zip(strings, inner_out):
            out.append(data[0]["generated_text"][len(s):]) # remove the input text from the output
        return out



import openai

class OpenAI():
    def __init__(self, engine="davinci-msft", max_tokens=50):
        self.engine = engine
        self.max_tokens = max_tokens

    def __call__(self, prompts, num_reps=1, temperature=0.9):
        response = openai.Completion.create(
            engine=self.engine, prompt=[p["prompt"] for p in prompts], max_tokens=self.max_tokens,
            temperature=temperature, n=num_reps, stop="\n"
        )
        return [choice["text"] for choice in response["choices"]]